#ifndef DUNE_TECTONIC_GLOBAL_LAURSEN_NONLINEARITY_HH
#define DUNE_TECTONIC_GLOBAL_LAURSEN_NONLINEARITY_HH

#include <vector>

#include <dune/common/fvector.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include "globalnonlinearity.hh"
#include "localnonlinearity.hh"
#include "nicefunction.hh"

namespace Dune {
template <int dim, class OuterFunctionType,
          class VectorType = BlockVector<FieldVector<double, dim>>,
          class MatrixType = BCRSMatrix<FieldMatrix<double, dim, dim>>>
class GlobalLaursenNonlinearity
    : public GlobalNonlinearity<dim, VectorType, MatrixType> {
public:
  GlobalLaursenNonlinearity(
      shared_ptr<BlockVector<FieldVector<double, 1>> const> mu,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> normalStress,
      shared_ptr<BlockVector<FieldVector<double, 1>> const> nodalIntegrals)
      : mu(mu), normalStress(normalStress), nodalIntegrals(nodalIntegrals) {}

  /*
    Return a restriction of the outer function to the i'th node. If
    mu and sigma_n denote the coefficient of friction and the normal
    stress, respectively, at the i'th node, this function is given
    by

    sigma_n [(1/eta \bar Gamma)* + mu id]

    TODO: We chose Gamma = id, so that (\bar Gamma)* = \Gamma^{-1}
    = id^{-1} = id. The factor 1/eta cancels in this special case, leaving us
    with

    sigma_n [id + mu id] = sigma_n (1 + mu) id
  */
  virtual shared_ptr<LocalNonlinearity<dim> const> restriction(int i) const {
    double coefficient = (*nodalIntegrals)[i][0];
    coefficient *= (*normalStress)[i];
    coefficient *= 1 + (*mu)[i];

    return make_shared<LocalNonlinearity<dim> const>(
        make_shared<OuterFunctionType const>(coefficient));
  }

private:
  // TODO: If we're clever, we only store one vector with the precomputed
  // results
  shared_ptr<BlockVector<FieldVector<double, 1>> const> mu;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> normalStress;
  shared_ptr<BlockVector<FieldVector<double, 1>> const> nodalIntegrals;
};
}
#endif
