// Based on dune/tnnmg/problem-classes/blocknonlineargsproblem.hh

#ifndef MY_BLOCK_PROBLEM_HH
#define MY_BLOCK_PROBLEM_HH

#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>

#include <dune/tnnmg/problem-classes/bisection.hh>
#include <dune/tnnmg/problem-classes/onedconvexfunction.hh>

#include "localnonlinearity.hh"
#include "mydirectionalconvexfunction.hh"
#include "samplefunctional.hh"

/** \brief Base class for problems where each block can be solved with a
 * modified gradient method */
template <class MyConvexProblemTypeTEMPLATE> class MyBlockProblem {
public:
  typedef MyConvexProblemTypeTEMPLATE MyConvexProblemType;
  typedef typename MyConvexProblemType::VectorType VectorType;
  typedef typename MyConvexProblemType::MatrixType MatrixType;
  typedef typename MyConvexProblemType::LocalVectorType LocalVectorType;
  typedef typename MyConvexProblemType::LocalMatrixType LocalMatrixType;

  int static const block_size = MyConvexProblemType::block_size;
  int static const coarse_block_size = block_size;

  /** \brief Solves one local system using a modified gradient method */
  class IterateObject;

  struct Linearization {
    static const int block_size = coarse_block_size;

    typedef typename MyBlockProblem<
        MyConvexProblemTypeTEMPLATE>::LocalMatrixType LocalMatrixType;
    typedef Dune::BCRSMatrix<typename Linearization::LocalMatrixType>
    MatrixType;
    typedef Dune::BlockVector<
        Dune::FieldVector<double, Linearization::block_size>> VectorType;
    typedef Dune::BitSetVector<Linearization::block_size> BitVectorType;
    typename Linearization::MatrixType A;
    typename Linearization::VectorType b;
    typename Linearization::BitVectorType ignore;

    Dune::BitSetVector<Linearization::block_size> truncation;
  };

  MyBlockProblem(Dune::ParameterTree const &parset,
                 MyConvexProblemType &problem)
      : parset(parset), problem(problem) {
    bisection = Bisection(
        0.0, // acceptError: Stop if the search interval has
             // become smaller than this number
        1.0, // acceptFactor: ?
        parset.get<double>("bisection.requiredResidual"), true, // fastQuadratic
        0); // safety: acceptance factor for inexact minimization
  }

  std::string virtual getOutput(bool header = false) const {
    // TODO: implement (not urgent)
    return std::string();
  }

  void projectCoarseCorrection(VectorType const &u,
                               typename Linearization::VectorType const &v,
                               VectorType &projected_v,
                               Linearization const &linearization) const {
    // TODO: implement (not urgent)
    projected_v = v;
  };

  double computeDampingParameter(VectorType const &u,
                                 VectorType const &projected_v) const {
    VectorType const v = projected_v;

    VectorType tmp = problem.f;
    problem.A.mmv(u, tmp);
    double const localb = tmp * v;

    problem.A.mv(v, tmp);
    double const localA = tmp * v;

    /*
      1/2 <A(u + hv),u + hv> - <b, u + hv>
      = 1/2 <Av,v> h^2 - <b - Au, v> h + const.

      localA = <Av,v>
      localb = <b - Au, v>
    */
    MyDirectionalConvexFunction<Dune::GlobalNonlinearity<block_size>> psi(
        localA, localb, problem.phi, u, v);

    int bisectionsteps = 0;
    Bisection bisection(0.0, 1.0, 1e-12, true, 0);            // TODO
    return bisection.minimize(psi, 0.0, 0.0, bisectionsteps); // TODO
  }

  void assembleTruncate(VectorType const &u, Linearization &linearization,
                        Dune::BitSetVector<block_size> const &ignore) const {
    // we can just copy the ignore information
    linearization.ignore = ignore;

    // determine truncation pattern
    linearization.truncation.resize(u.size());
    linearization.truncation.unsetAll();
    for (int i = 0; i < u.size(); ++i)
      // TODO: should ignoreNodes be truncated as well?
      if (problem.phi.regularity(i, u[i]) > 1e8) // TODO
        linearization.truncation[i] = true;

    // construct sparsity pattern for linearization
    Dune::MatrixIndexSet indices(problem.A.N(), problem.A.M());
    indices.import(problem.A);
    problem.phi.addHessianIndices(indices);

    // construct matrix from pattern and initialize it
    indices.exportIdx(linearization.A);
    linearization.A = 0.0;

    // compute quadratic part of hessian (linearization.A += problem.A)
    for (int i = 0; i < problem.A.N(); ++i) {
      typename MatrixType::row_type::ConstIterator it = problem.A[i].begin();
      typename MatrixType::row_type::ConstIterator end = problem.A[i].end();
      for (; it != end; ++it)
        StaticMatrix::axpy(linearization.A[i][it.index()], 1.0, *it);
    }

    // compute nonlinearity part of hessian
    problem.phi.addHessian(u, linearization.A);

    // compute quadratic part of gradient
    linearization.b.resize(u.size());
    linearization.b = 0.0;
    problem.A.template umv<VectorType, VectorType>(u, linearization.b);
    linearization.b -= problem.f;

    // compute nonlinearity part of gradient
    problem.phi.addGradient(u, linearization.b);

    // -grad is needed for Newton step
    linearization.b *= -1.0;

    // apply truncation to system
    typename Linearization::MatrixType::row_type::Iterator it;
    typename Linearization::MatrixType::row_type::Iterator end;
    for (int row = 0; row < linearization.A.N(); ++row) {
      it = linearization.A[row].begin();
      end = linearization.A[row].end();
      for (; it != end; ++it) {
        int const col = it.index();
        for (int i = 0; i < it->N(); ++i) {
          typename Linearization::MatrixType::block_type::row_type::Iterator
          blockIt = (*it)[i].begin();
          typename Linearization::MatrixType::block_type::row_type::
              Iterator const blockEnd = (*it)[i].end();
          for (; blockIt != blockEnd; ++blockIt)
            if (linearization.truncation[row][i] or linearization
                    .truncation[col][blockIt.index()])
              *blockIt = 0.0;
        }
      }

      for (int j = 0; j < block_size; ++j)
        if (linearization.truncation[row][j])
          linearization.b[row][j] = 0.0;
    }

    // for(int j=0; j<block_size; ++j)
    //   outStream << std::setw(9) << linearization.truncation.countmasked(j);
  }

  /** \brief Constructs and returns an iterate object */
  IterateObject getIterateObject() {
    return IterateObject(parset, bisection, problem);
  }

private:
  // problem data
  MyConvexProblemType &problem;

  // commonly used minimization stuff
  Bisection bisection;

  Dune::ParameterTree const &parset;
};

/** \brief Solves one local system using a scalar Gauss-Seidel method */
template <class MyConvexProblemTypeTEMPLATE>
class MyBlockProblem<MyConvexProblemTypeTEMPLATE>::IterateObject {
  friend class MyBlockProblem;

protected:
  /** \brief Constructor, protected so only friends can instantiate it
   * \param bisection The class used to do a scalar bisection
   * \param problem The problem including quadratic part and nonlinear part
   */
  IterateObject(Dune::ParameterTree const &parset, Bisection const &bisection,
                MyConvexProblemType &problem)
      : parset(parset), problem(problem), bisection(bisection) {}

public:
  /** \brief Set the current iterate */
  void setIterate(VectorType &u) {
    this->u = u;
    return;
  }

  /** \brief Update the i-th block of the current iterate */
  void updateIterate(LocalVectorType const &ui, int i) {
    u[i] = ui;
    return;
  }

  /** \brief Minimise a local problem using a modified gradient method
   * \param[out] ui The solution
   * \param m Block number
   * \param ignore Set of degrees of freedom to leave untouched
   */
  void solveLocalProblem(
      LocalVectorType &ui, int m,
      const typename Dune::BitSetVector<block_size>::const_reference ignore) {
    {
      int ignore_component =
          block_size; // Special value that indicates nothing should be ignored
      switch (ignore.count()) {
        case 0: // Full problem
          break;
        case 1: // 1 Dimension is fixed
          assert(
              ignore[1]); // Only the second component is allowed to stay fixed
          ignore_component = 1;
          break;
        case block_size: // Ignore the whole node
          return;
        default:
          assert(false);
      }

      LocalMatrixType const *localA = NULL;
      LocalVectorType localb(problem.f[m]);

      typename MatrixType::row_type::ConstIterator it;
      typename MatrixType::row_type::ConstIterator end = problem.A[m].end();
      for (it = problem.A[m].begin(); it != end; ++it) {
        int const j = it.index();
        if (j == m)
          localA = &(*it); // localA = A[m][m]
        else
          it->mmv(u[j], localb); // localb -= A[m][j] * u[j]
      }
      assert(localA != NULL);

      auto phi = problem.phi.restriction(m);
      Dune::SampleFunctional<block_size> localJ(*localA, localb, phi,
                                                ignore_component);

      LocalVectorType correction;
      Dune::minimise(localJ, ui, parset.get<size_t>("localsolver.steps"),
                     bisection);
    }
  }

private:
  Dune::ParameterTree const &parset;

  // problem data
  MyConvexProblemType &problem;

  // commonly used minimization stuff
  Bisection bisection;

  // state data for smoothing procedure used by:
  // setIterate, updateIterate, solveLocalProblem
  VectorType u;
};

#endif
